package org.netkernel.lang.dpml.ast.impl;

import org.netkernel.lang.dpml.ast.*;
import org.netkernel.lang.dpml.endpoint.CompilerScope;
import org.netkernel.lang.dpml.endpoint.DPMLUtils;
import org.netkernel.lang.dpml.endpoint.RuntimeState;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.NKFException;

import java.util.*;


public class Closure extends VariableImpl implements IScope, IValueVariable
{
	private Map mAssignments;
	private List mImports = new ArrayList();
	private CompilerScope mCompilerScope;
	
	public static final int PREC_LOW=-1;
	public static final int PREC_SAME=0;
	public static final int PREC_HIGH=1;
	
	public static final String RESPONSE="response";
	
	public Closure(String aName, IScope aScope, ILocation aLocation, CompilerScope aCompilerScope)
	{	super(aName,aScope,aLocation);
		mAssignments = new HashMap();
		mCompilerScope=aCompilerScope;
	}

	public IScope getParent()
	{	return getScope();
	}
	
	public void addAssignment(IVariable aVariable, boolean aReplace) throws NKFException
	{	IVariable existingAssignment  =(IVariable)mAssignments.put(aVariable.getName(),aVariable);
		if (existingAssignment!=null && !aReplace)
		{	throw new DuplicateAssignmentException();
		}
	}

	public void setResponse(IVariable aResponse)
	{	mAssignments.put(RESPONSE,aResponse);
	}
	
	public IVariable resolveAssignment(String aName, RuntimeState aState)
	{	IVariable result=(IVariable)mAssignments.get(aName);
		if (result==null && this.getParent()!=null)
		{	result=this.getParent().resolveAssignment(aName,aState);
		}
		return result;
	}
	
	public Collection<IVariable> getVariables()
	{	return mAssignments.values();
	}
	
	
	public IValue getValue(INKFRequestContext aContext, RuntimeState aState, Class aRepresentationClass) throws NKFException
	{
		RuntimeState state=new RuntimeState(aState,this);
		
		//find response assignment or first and only assignment
		IVariable resp = (IVariable)mAssignments.get(RESPONSE);
		if (resp==null)
		{	//throw aContext.createFormattedException("EX_DPML_RUNTIME","MSG_PXML_NO_RESPONSE", DPMLUtils.formatLocationMessage(this,aContext),null);
			
			if (mAssignments.size()==1)
			{	resp= (IVariable)mAssignments.values().iterator().next();
				if (!resp.getName().startsWith(VariableImpl.ANON_PREFIX))
				{	resp=null;
				}
			}
			if (resp==null)
			{	throw aContext.createFormattedException("EX_DPML_RUNTIME","MSG_PXML_NO_RESPONSE", DPMLUtils.formatLocationMessage(this,aContext),null);
			}
		}				

		resp = DPMLUtils.dereference(resp,state,aContext);
		IValue result=null;
		if (resp instanceof IIdentifierVariable)
		{	result=DPMLUtils.requestIdentifier((IIdentifierVariable)resp,aContext,state,aRepresentationClass);
		}
		else if (resp instanceof IValueVariable)
		{	result= ((IValueVariable)resp).getValue(aContext,state,aRepresentationClass);
		}
		return result;
	}
	
	public void addImport(IScope aImport, int aPrecedence)
	{	ImportDef importDef = new ImportDef(aImport,aPrecedence);
		mImports.add(importDef);
	}
	
	public void close(INKFRequestContext aContext) throws NKFException
	{
		if (mImports!=null)
		{
			for (Iterator i=mImports.iterator(); i.hasNext(); )
			{	ImportDef importDef = (ImportDef)i.next();
				for (Iterator j=importDef.mRootScope.getVariables().iterator(); j.hasNext(); )
				{	IVariable var = (IVariable)j.next();
					boolean existingVar = mAssignments.get(var.getName())!=null;
					if (!existingVar)
					{	addAssignment(var,false);
					}
					else
					{	switch (importDef.mPrecedence)
						{	case PREC_SAME:
								throw aContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_IMPORT_CONFLICT",null,null,
										var.getName(), var.getLocation().getSourceIdentity() );
							case PREC_LOW:
								break; // ignore the imported defn
							case PREC_HIGH:
								addAssignment(var,true); // replace existing
								break;
						}
					}
				}
			}
			mImports=null;
		}
		
		
	}
	
	public CompilerScope getCompilerScope()
	{	return mCompilerScope;
	}
	
	public String toString()
	{	return "Closure["+getName()+"]";
	}
	
	private static class ImportDef
	{
		public final IScope mRootScope;
		public final int mPrecedence;
		
		public ImportDef(IScope aRootScope, int aPrecedence)
		{	mRootScope=aRootScope;
			mPrecedence=aPrecedence;
		}
	}

}
