package org.netkernel.lang.dpml.endpoint;

import org.netkernel.lang.dpml.ast.IArgument;
import org.netkernel.lang.dpml.ast.ILocation;
import org.netkernel.lang.dpml.ast.IRequestHeader;
import org.netkernel.lang.dpml.ast.IScope;
import org.netkernel.lang.dpml.ast.IVariable;
import org.netkernel.lang.dpml.ast.impl.*;
import org.netkernel.lang.dpml.endpoint.CompilerScope.RequestAlias;
import org.netkernel.lang.dpml.representation.DPMLRepresentation;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.module.standard.endpoint.StandardTransreptorImpl;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.xml.representation.IRepStAX;
import org.netkernel.xml.util.XMLUtils;
import org.w3c.dom.Document;

import javax.xml.stream.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

public class DPMLCompiler extends StandardTransreptorImpl
{
	private final CompilerScope mRootScope;
	
	public DPMLCompiler()
	{	declareThreadSafe();
		declareToRepresentation(DPMLRepresentation.class);
		declareDescription("Compile XML Syntax to DPML");
		mRootScope=new CompilerScope(null);
		mRootScope.declareAlias(new RequestAlias("if","meta:dpml:if"));
		mRootScope.declareAlias(new RequestAlias("throw","meta:dpml:throw"));
		mRootScope.declareAlias(new RequestAlias("exception","meta:dpml:exception"));
		mRootScope.declareAlias(new RequestAlias("modify-response","meta:dpml:modify-response"));
		mRootScope.declareAlias(new RequestAlias("log","meta:dpml:log"));
		mRootScope.declareAlias(new RequestAlias("switch","meta:dpml:switch"));
	}

public void onTransrept(INKFRequestContext aContext) throws Exception
	{		
		//String source=aContext.sourcePrimary(String.class);
		//System.out.println(source);
		IRepStAX stax = (IRepStAX)aContext.sourcePrimary(IRepStAX.class);
		XMLStreamReader r = stax.getXMLStreamReader();
		String requestId=aContext.getThisRequest().getIdentifier();
		aContext.setCWU(requestId);
		
		DPMLBuilder b = new DPMLBuilder(r,requestId,aContext,mRootScope);
		DPMLRepresentation representation = new DPMLRepresentation(b.getRootScope(),b.getVariables(),requestId);
		aContext.createResponseFrom(representation);
	}
	
	private static class DPMLBuilder
	{
		private static final String TAG_CLOSURE="closure";
		private static final String TAG_SEQUENCE="sequence";
		private static final String TAG_ASSIGNMENT="assignment";
		private static final String TAG_REQUEST="request";
		private static final String TAG_BASE="identifier";
		private static final String TAG_VERB="verb";
		private static final String TAG_ARGUMENT="argument";
		private static final String TAG_ARGNAME="name";
		private static final String TAG_REPRESENTATION="representation";
		private static final String TAG_ARGMETHOD="method";
		private static final String TAG_LITERAL="literal";
		private static final String TAG_LITERALTYPE="type";
		private static final String TAG_DEREFERENCE="dereference";
		private static final String TAG_IMPORT="import";
		private static final String TAG_IMPORTPREC="precedence";
		private static final String TAG_COMMENT="comment";
		private static final String TAG_HEADER="header";
		private static final String TAG_HEADERNAME="name";
		private static final String TAG_HEADERSTICKY="sticky";
		private static final String TAG_VARARGS="varargs";

		private static final String TAG_DECLARE_TAG="declare-tag";
		private static final String TAG_DECLARE_TAG_NAME="name";
		private static final String TAG_DECLARE_TAG_ID="identifier";

		
		private static final String LITERAL_TYPE_URI="identifier";
		private static final String LITERAL_TYPE_ABSOLUTE="absolute-identifier";

		private static final String ARG_METHOD_EAGER="eager";
		private static final String ARG_METHOD_LAZY="lazy";
		private static final String ARG_METHOD_IDENTITY="identity";
		
		private static final String PREC_LOW="lower";
		private static final String PREC_SAME="same";
		private static final String PREC_HIGH="higher";
		
		private static final String THIS_PREFIX="this:";
		private static final String ARG_PREFIX="arg:";
		
		private AtomicInteger mUniqueId=new AtomicInteger();
		
		private XMLStreamReader mReader;
		private IScope mRoot;
		//private Map mVariablesById;
		private List mVariables;
		private INKFRequestContext mContext;
		private String mRequestId;
		
		public DPMLBuilder(XMLStreamReader aReader, String aRequestId, INKFRequestContext aContext, CompilerScope aRootScope) throws Exception
		{	mReader=aReader;
			mRequestId=aRequestId;
			mContext=aContext;
			//mVariablesById = new HashMap();
			mVariables=new ArrayList();
			while(mReader.getEventType()!=XMLStreamReader.START_ELEMENT)
			{	mReader.next();
			}
			String n=mReader.getLocalName();
			if (n.equals(TAG_CLOSURE))
			{	parseClosure(null,aRootScope,false);
			}
			else if (n.equals(TAG_SEQUENCE))
			{	parseClosure(null,aRootScope,true);
			}
			else
			{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_ELEMENT",formatLocationMessage(),null,n);
			}
			
			// finalize any variables that need stuff done after all context is known
			for (Iterator i=mVariables.iterator(); i.hasNext(); )
			{	IVariable var = (IVariable)i.next();
				var.close(aContext);
			}
		}
		
		private String getAssignment() throws Exception
		{	String assignment = mReader.getAttributeValue(null,TAG_ASSIGNMENT);
			if (assignment==null)
			{	assignment=VariableImpl.ANON_PREFIX+Integer.toHexString(mUniqueId.getAndIncrement());
			}
			else if (assignment.startsWith(VariableImpl.ANON_PREFIX))
			{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_BAD_ASSIGMENT",formatLocationMessage(),null,assignment);
			}
			return assignment;
		}
		
		
		/** remember variables so we can do any initialisation once whole idoc is parsed */
		private void putVariable(IVariable aVariable)
		{	//mVariablesById.put(aVariable.getId(),aVariable);
			mVariables.add(aVariable);
		}
		
		public IScope getRootScope()
		{	return mRoot;
		}
		/*
		public Map getVariables()
		{	return mVariablesById;
		}
		*/
		public List getVariables()
		{	return mVariables;
		}
		
		private String toStringCursor()
		{	String type;
			switch (mReader.getEventType())
			{	case XMLStreamReader.ATTRIBUTE: type="ATTRIBUTE"; break;
				case XMLStreamReader.CDATA: type="CDATA"; break;
				case XMLStreamReader.CHARACTERS: type="CHARACTERS("+mReader.getText()+")"; break;
				case XMLStreamReader.COMMENT: type="COMMENT"; break;
				case XMLStreamReader.DTD: type="DTD"; break;
				case XMLStreamReader.END_DOCUMENT: type="END_DOCUMENT"; break;
				case XMLStreamReader.END_ELEMENT: type="END_ELEMENT("+mReader.getLocalName()+")"; break;
				case XMLStreamReader.ENTITY_DECLARATION: type="ENTITY_DECLARATION"; break;
				case XMLStreamReader.ENTITY_REFERENCE: type="ENTITY_REFERENCE"; break;
				case XMLStreamReader.NAMESPACE: type="NAMESPACE"; break;
				case XMLStreamReader.NOTATION_DECLARATION: type="NOTATION_DECLARATION"; break;
				case XMLStreamReader.PROCESSING_INSTRUCTION: type="PROCESSING_INSTRUCTION"; break;
				case XMLStreamReader.SPACE: type="SPACE"; break;
				case XMLStreamReader.START_DOCUMENT: type="START_DOCUMENT"; break;
				case XMLStreamReader.START_ELEMENT: type="START_ELEMENT("+mReader.getLocalName()+")"; break;
				default: type="UNKNOWN"; break;
			}
			return type;	
		}
		
		private static final Object[] sLocParams = new Object[3];
		private String formatLocationMessage()
		{
			Location loc=mReader.getLocation();
			String msg;
			synchronized(sLocParams)
			{	sLocParams[0]=mRequestId;
				sLocParams[1]=new Integer(loc.getLineNumber());
				sLocParams[2]=new Integer(loc.getColumnNumber());
				msg=mContext.formatMessage("MSG_PXML_LOCATION",sLocParams);
			}
			return msg;
		}
		
		private ILocation getLocation()
		{	Location loc=mReader.getLocation();
			return new LocationImpl(mRequestId,loc.getLineNumber(),loc.getColumnNumber());
		}
		
		private IVariable parseClosure(IScope aParent, CompilerScope aCompilerParent, boolean aSequential) throws Exception
		{
			CompilerScope compilerScope = new CompilerScope(aCompilerParent);
			String assignment = getAssignment();
			IScope result;
			if (aSequential)
			{	result =new Sequence(assignment,aParent,getLocation(),compilerScope);
			}
			else
			{	result =new Closure(assignment,aParent,getLocation(),compilerScope);
			}
			putVariable(result);
			if (aParent==null)
			{	mRoot=result;
			}
			
			IVariable defResponse=null;
			int subElementCount=0;
			while (mReader.hasNext())
			{	int event=mReader.next();
			
				if (event==XMLStreamReader.START_ELEMENT)
				{	String n=mReader.getLocalName();
					RequestAlias ra;
					IVariable var;
					if (n.equals(TAG_CLOSURE))
					{	var = parseClosure(result,compilerScope,false);
					}
					else if (n.equals(TAG_SEQUENCE))
					{	var = parseClosure(result,compilerScope,true);
					}
					else if (n.equals(TAG_REQUEST))
					{	var = parseRequest(result,compilerScope);
					}					
					else if (n.equals(TAG_LITERAL))
					{	var= parseLiteral(result);
					}
					else if (n.equals(TAG_DEREFERENCE))
					{	var = parseDereference(result,compilerScope);
					}
					else if (n.equals(TAG_IMPORT))
					{	if (aSequential)
						{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_IMPORT_IN_SEQ",formatLocationMessage(),null);
						}
						parseImport(result,compilerScope);
						//TODO need to deal with subElementCount
						var=null;
					}
					else if (n.equals(TAG_COMMENT))
					{	parseComment();
						var=null;
					}
					else if (n.equals(TAG_DECLARE_TAG))
					{	parseDeclareTag(compilerScope);
						var=null;
					}
					else if ( (ra=compilerScope.lookupTag(n))!=null)
					{	var=parseTagAlias(ra,result,compilerScope);
					}
					else
					{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_ELEMENT",formatLocationMessage(),null,n);
					}
					if (var!=null)
					{	putVariable(var);
						try
						{	result.addAssignment(var,false);
						}
						catch (DuplicateAssignmentException e)
						{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_DUPLICATE_ASSIGNMENT",formatLocationMessage(),null,var.getName());
						}
						subElementCount++;
						if (defResponse==null)
						{	defResponse=var;
						}
					}
				
				}
				else if (event==XMLStreamReader.END_ELEMENT)
				{	String n=mReader.getLocalName();
					if (n.equals(TAG_CLOSURE) || n.equals(TAG_SEQUENCE))
					{	break;
					}
				}
			}		
			return result;
		}
		
		private IVariable parseTagAlias(RequestAlias aRA,IScope aScope, CompilerScope aCompilerScope) throws Exception
		{
			String assignment = getAssignment();
			List arguments = new ArrayList(8);
			while (mReader.hasNext())
			{	int event=mReader.next();
				if (event==XMLStreamReader.START_ELEMENT)
				{	String argName=mReader.getLocalName();
					//if (!aRA.supportsArgument(argName))
					//{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_ARG",formatLocationMessage(),null,argName,aRA.getTag());
					//}
					int argMethod = parseMethod();
					mReader.next();
					IVariable argVar = parseArgument(aScope,aCompilerScope);
					IArgument arg = new ArgumentImpl(argName,argMethod,argVar);
					arguments.add(arg);
					parseTextUntilEndOf(argName);
				}
				else if (event==XMLStreamReader.END_ELEMENT && mReader.getLocalName().equals(aRA.getTag()))
				{	break;
				}
			}
			
			String identifier=aRA.getIdentifier();
			return new SubRequest(assignment,aScope,getLocation(),identifier,INKFRequestReadOnly.VERB_SOURCE,arguments,Collections.EMPTY_LIST,null,false);
		}
		
		private IVariable parseRequest(IScope aScope, CompilerScope aCompilerScope) throws Exception
		{	String assignment = getAssignment();
			String base=null;
			int verb=INKFRequestReadOnly.VERB_SOURCE;
			String verbString="SOURCE";
			String repName=null;
			List arguments = new ArrayList(8);
			List headers = new ArrayList(4);
			boolean varargs=false;
			while (mReader.hasNext())
			{	int event=mReader.next();
				if (event==XMLStreamReader.START_ELEMENT)
				{	String name=mReader.getLocalName();
					if (name.equals(TAG_BASE))
					{	mReader.next();
						base = parseTextUntilEndOf(TAG_BASE);
						//base=DPMLSubrequestContext.escapeArgIdentifier(base);
					}
					else if (name.equals(TAG_ARGUMENT))
					{	String argName=mReader.getAttributeValue(null,TAG_ARGNAME);
						int argMethod = parseMethod();
						mReader.next();
						IVariable argVar = parseArgument(aScope,aCompilerScope);
						IArgument arg = new ArgumentImpl(argName,argMethod,argVar);
						arguments.add(arg);
					}
					else if (name.equals(TAG_VERB))
					{	mReader.next();
						verbString = parseTextUntilEndOf(TAG_VERB);
						verb=RequestFactory.parseVerb(verbString);
						if (verb==0)
						{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_VERB",formatLocationMessage(),null);
						}
					}
					else if  (name.equals(TAG_REPRESENTATION))
					{	mReader.next();
						repName = parseTextUntilEndOf(TAG_REPRESENTATION);
					}
					else if (name.equals(TAG_COMMENT))
					{	parseComment();
					}
					else if (name.equals(TAG_HEADER))
					{	String headerName=mReader.getAttributeValue(null,TAG_HEADERNAME);
						boolean headerSticky = parseSticky();
						mReader.next();
						IVariable headerVar = parseArgument(aScope,aCompilerScope);
						IRequestHeader header = new HeaderImpl(headerName,headerSticky,headerVar);
						headers.add(header);
					}
					else if (name.equals(TAG_VARARGS))
					{	varargs=true;
					}
					else
					{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_ELEMENT",formatLocationMessage(),null,name);
					}
				}
				else if (event==XMLStreamReader.END_ELEMENT && mReader.getLocalName().equals(TAG_REQUEST))
				{	break;
				}
			}
			if (arguments.size()==0)
			{	//same memory if we have no arguments
				arguments=Collections.EMPTY_LIST;
			}
			if (base!=null)
			{	if (base.startsWith(this.THIS_PREFIX))
				{	String toCall=base.substring(THIS_PREFIX.length());
					return new AugmentedClosure(assignment, aScope,getLocation(), toCall, verbString, arguments);
				}
				else if (base.startsWith(this.ARG_PREFIX))
				{	if (arguments.size()>0)
					{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_BAD_ARG_OP",formatLocationMessage(),null);
					}
					String argName=base.substring(ARG_PREFIX.length());
					return new ArgumentReference(assignment,aScope,getLocation(),argName,verb);
				}
				else
				{	return new SubRequest(assignment,aScope,getLocation(),base,verb,arguments,headers,repName,varargs);
				}
			}
			else
			{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_REQUEST_TYPE",formatLocationMessage(),null);
			}
			
		}
		
		private int parseMethod() throws Exception
		{	int result;
			String method=mReader.getAttributeValue(null,TAG_ARGMETHOD);
			if (method==null)
			{	result = IArgument.METHOD_DEFAULT;
			}
			else if (ARG_METHOD_EAGER.equals(method))
			{	result = IArgument.METHOD_EAGER;
			}
			else if (ARG_METHOD_LAZY.equals(method))
			{	result= IArgument.METHOD_LAZY;
			}
			else if (ARG_METHOD_IDENTITY.equals(method))
			{	result=IArgument.METHOD_IDENTITY;
			}
			else
			{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_METHOD",formatLocationMessage(),null);
			}
			return result;
		}
		
		private boolean parseSticky() throws Exception
		{	boolean result;
			String sticky=mReader.getAttributeValue(null,TAG_HEADERSTICKY);
			if (sticky==null)
			{	result =false;
			}
			else
			{	sticky=sticky.toLowerCase();	
				if ("true".equals(sticky))
				{	result = true;
				}
				else if ("false".equals(sticky))
				{	result=false;
				}
				else
				{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_STICKY",formatLocationMessage(),null);
				}
			}
			return result;
		}
		

		private String parseTextUntilEndOf(String aTagName) throws Exception
		{	
			String result=null;
			do
			{	int event=mReader.getEventType();
				switch(event)
				{	case XMLStreamReader.CHARACTERS:
					{	String t=mReader.getText();
						if (result==null)
						{	result=t;
						}
						else
						{	result=result+t;
						}
						break;
					}
					case XMLStreamReader.COMMENT:
						break;
					case XMLStreamReader.SPACE:
						break;
					case XMLStreamReader.END_ELEMENT:
						return (result==null)?"":result.trim();
					case XMLStreamReader.START_ELEMENT:
						throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNKNOWN_ELEMENT",formatLocationMessage(),null,mReader.getLocalName());
					default:
						throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_EXPECTED_END",formatLocationMessage(),null,aTagName,toStringCursor());
				}
				mReader.next();
			}
			while (true);
		}
		
		private IVariable parseArgument(IScope aScope, CompilerScope aCompilerScope) throws Exception
		{	IVariable result=null;
		
			while(true)
			{	int type=mReader.getEventType();
				if (type==XMLStreamReader.START_ELEMENT)
				{	String tag=mReader.getLocalName();
					RequestAlias ra;
					IVariable innerResult=null;
					if (result==null)
					{	if (TAG_CLOSURE.equals(tag))
						{	innerResult = parseClosure(aScope,aCompilerScope,false);
						}
						else if (TAG_SEQUENCE.equals(tag))
						{	innerResult = parseClosure(aScope,aCompilerScope,true);
						}
						else if (TAG_REQUEST.equals(tag))
						{	innerResult = parseRequest(aScope, aCompilerScope);
						}
						else if (TAG_LITERAL.equals(tag))
						{	innerResult = parseLiteral(aScope);
						}
						else if (TAG_DEREFERENCE.equals(tag))
						{	innerResult = parseDereference(aScope,aCompilerScope);
						}
						else if (tag.equals(TAG_COMMENT))
						{	parseComment();
						}
						else if ( (ra=aCompilerScope.lookupTag(tag))!=null)
						{	innerResult=parseTagAlias(ra,aScope,aCompilerScope);
						}
						else
						{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNEXPECTED_ARG_CONTENT",formatLocationMessage(),null,tag);
						}
						if (innerResult!=null)
						{	String name=innerResult.getName();
							if (name.startsWith(VariableImpl.ANON_PREFIX) /*|| name.equals(Closure.RESPONSE)*/)
							{
								if (result==null)
								{	putVariable(innerResult);
									result=innerResult;
								}
								else
								{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_MULTIPLE_INSIDE_ARG",formatLocationMessage(),null,tag);
								}
							}
							else
							{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_ASSIGNMENT_ON_ARG",formatLocationMessage(),null,innerResult.getName());
							}
						}
					}
					else
					{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_UNEXPECTED_ARG_CONTENT",formatLocationMessage(),null,toStringCursor());
					}
				}
				else if (type==XMLStreamReader.CHARACTERS)
				{	String text = mReader.getText();
					if (text.trim().length()>0)
					{	String identifier = parseTextUntilEndOf(TAG_ARGUMENT).trim();
						result = createReference(null,identifier,false,aScope);
						break;
					}
					else
					{	mReader.next();
					}
				}
				else if (type==XMLStreamReader.END_ELEMENT)
				{	mReader.next();
					break;
				}
				else
				{	mReader.next();
				}
			}
			return result;
		}
		
		private IVariable createReference(String aName,String aReference, boolean aAbsolute, IScope aScope) throws Exception
		{	IVariable result;
			if (aReference.startsWith(THIS_PREFIX))
			{	String refName = aReference.substring(THIS_PREFIX.length());
				result = new AssignmentReference(aName,aScope,getLocation(),refName); 
			}
			else if (aReference.startsWith(ARG_PREFIX))
			{	String argName = aReference.substring(ARG_PREFIX.length());
				result = new ArgumentReference(aName,aScope,getLocation(),argName,INKFRequestReadOnly.VERB_SOURCE);
			}
			else
			{	result = new LiteralIndentifier(aName,aScope,getLocation(),aReference,aAbsolute);
			}
			return result;
		}

		private IVariable parseLiteral(IScope aScope) throws Exception
		{	IVariable result;
			
			String assignment = getAssignment();
			String type = mReader.getAttributeValue(null,TAG_LITERALTYPE);
			
			if (LITERAL_TYPE_URI.equals(type))
			{	mReader.next();
				String reference=parseTextUntilEndOf(TAG_LITERAL).trim();
				result = createReference(assignment,reference,false,aScope);
			}
			else if (LITERAL_TYPE_ABSOLUTE.equals(type))
			{	mReader.next();
				String reference=parseTextUntilEndOf(TAG_LITERAL).trim();
				result = createReference(assignment,reference,true,aScope);
			}
			else
			{	try
				{	XMLOutputFactory f=XMLUtils.getXMLOutputFactory();
					ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
					String encoding="UTF-8";
					XMLStreamWriter w = f.createXMLStreamWriter(baos,encoding);
					w.writeStartDocument(encoding,"1.0");
					parseXMLUntilEndOfElement(w);
					w.writeEndDocument();
					w.close();
					InputStream is = new ByteArrayInputStream(baos.toByteArray());
					Document d = org.netkernel.layer0.util.XMLUtils.parse(is);
					Object[] argTypePair=RequestBuilder.parseLiteral(d.getDocumentElement(),null,mContext);
					Object value=argTypePair[0];
					
					IResponseMeta meta = MetaImpl.NEVER_EXPIRED;
					result = new LiteralValue(assignment,aScope,getLocation(),value,meta);
				}
				catch (Exception e)
				{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_BAD_LITERAL",formatLocationMessage(),e);
				}
			}
			
			return result;
		}
		
		private void parseXMLUntilEndOfElement(XMLStreamWriter w) throws XMLStreamException, IOException
		{	int depth=0;
			while(true)
			{	
				int type=mReader.getEventType();
				if (type==XMLStreamReader.START_ELEMENT)
				{	depth++;
					
					String prefix=mReader.getPrefix();
					String local=mReader.getLocalName();
					String uri=mReader.getNamespaceURI();
					//Note the great API to XMLStreamWriter here!
					if (prefix!=null && !prefix.equals("") && uri!=null)
					{	w.writeStartElement(prefix,local,uri);
					}
					else if (uri!=null)
					{	w.writeStartElement(uri,local);
					}
					else
					{	w.writeStartElement(local);
					}
					for (int i=0; i<mReader.getNamespaceCount(); i++)
					{	prefix=mReader.getNamespacePrefix(i);
						uri=mReader.getNamespaceURI(i);
						w.writeNamespace(prefix,uri);
					}
					
					
					for (int i=0; i<mReader.getAttributeCount(); i++)
					{
						prefix=mReader.getAttributePrefix(i);
						local=mReader.getAttributeLocalName(i);
						uri=mReader.getAttributeNamespace(i);
						String value=mReader.getAttributeValue(i);
						if (prefix!=null && !prefix.equals("") && uri!=null)
						{	w.writeAttribute(prefix,uri,local,value);
						}
						else if (uri!=null)
						{	w.writeAttribute(uri,local,value);
						}
						else
						{	w.writeAttribute(local,value);
						}
					}
				}
				else if (type==XMLStreamReader.END_ELEMENT)
				{	depth--;
					w.writeEndElement();
					if (depth==0)
					{	break;
					}
				}
				else if (type==XMLStreamReader.CHARACTERS)
				{	w.writeCharacters(mReader.getText());
				}
				mReader.next();
			}
		}		
		
		private void parseComment() throws XMLStreamException, IOException
		{	int depth=0;
			while(true)
			{	int type=mReader.getEventType();
				if (type==XMLStreamReader.START_ELEMENT)
				{	depth++;
				}
				else if (type==XMLStreamReader.END_ELEMENT)
				{	depth--;
					if (depth==0)
					{	break;
					}
				}
				mReader.next();
			}
		}		
		
		private void parseDeclareTag(CompilerScope aCompilerScope) throws Exception
		{	
			String name = mReader.getAttributeValue(null,TAG_DECLARE_TAG_NAME);
			mReader.next();
			String identifier=parseTextUntilEndOf(TAG_DECLARE_TAG);			
			RequestAlias alias = new RequestAlias(name, identifier);
			aCompilerScope.declareAlias(alias);
			
		}
		
		
		
		
		private IVariable parseDereference(IScope aScope, CompilerScope aCompilerScope) throws Exception
		{	
			String assignment = getAssignment();
			mReader.next();
			IVariable inner = parseArgument(aScope, aCompilerScope);
			return new Dereference(assignment,aScope,getLocation(),inner);
		}
		
		private void parseImport(IScope aScope, CompilerScope aCompilerScope) throws Exception
		{
			String precedence = mReader.getAttributeValue(null,TAG_IMPORTPREC);
			if (precedence==null)
			{	precedence=PREC_SAME;
			}
			
			mReader.next();
			String importId=parseTextUntilEndOf(TAG_IMPORT);
			
			/*
			IIdentifier importURI=new SimpleIdentifierImpl(importId);
			if (importURI.isRelative())
			{	SimpleIdentifierImpl base = new SimpleIdentifierImpl(mRequestId);
				importURI = base.resolve(importURI);
			}
			*/
			
			DPMLRepresentation rep = (DPMLRepresentation)mContext.source(importId, DPMLRepresentation.class);
			mVariables.addAll(rep.getVariables());
			//add all variables from import into this scope
			int precInt;
			if (precedence.equals(PREC_LOW))
			{	precInt=Closure.PREC_LOW;
			}
			else if (precedence.equals(PREC_SAME))
			{	precInt=Closure.PREC_SAME;
			}
			else if (precedence.equals(PREC_HIGH))
			{	precInt=Closure.PREC_HIGH;
			}
			else
			{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_PREC_UNKNOWN",formatLocationMessage(),null);
			}
			IScope rootScope=rep.getRootScope();
			if (rootScope instanceof Sequence)
			{	throw mContext.createFormattedException("EX_PXML_PARSE","MSG_PXML_IMPORT_OF_SEQ",formatLocationMessage(),null);
			}
			aScope.addImport(rootScope,precInt);
			
			//copy over any tag alias from root scope of import
			for (RequestAlias alias: rootScope.getCompilerScope().getAliases())
			{	aCompilerScope.declareAlias(alias);
			}

		}
		
				
	}

}
